"""A process function for determing offensive efficiency of entities."""

import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .identifier import Identifier

OFFENSIVE_EFFICIENCY_COLUMN = "offensiveefficiency"


def offensive_efficiency_process(
    df: pd.DataFrame, identifiers: list[Identifier]
) -> pd.DataFrame:
    """Process a dataframe for offensive efficiency."""
    tqdm.pandas(desc="Offensive Efficiency Features")

    def record_offensive_efficiency(row: pd.Series) -> pd.Series:
        nonlocal identifiers
        for identifier in identifiers:
            if identifier.field_goals_column is None:
                continue
            field_goals_value = row[identifier.field_goals_column]
            if pd.isnull(field_goals_value):
                continue
            field_goals = float(field_goals_value)
            if identifier.assists_column is None:
                continue
            assists_value = row[identifier.assists_column]
            if pd.isnull(assists_value):
                continue
            assists = float(assists_value)
            if identifier.field_goals_attempted_column is None:
                continue
            field_goals_attempted_value = row[identifier.field_goals_attempted_column]
            if pd.isnull(field_goals_attempted_value):
                continue
            field_goals_attempted = float(row[identifier.field_goals_attempted_column])
            if identifier.offensive_rebounds_column is None:
                continue
            offensive_rebounds_value = row[identifier.offensive_rebounds_column]
            if pd.isnull(offensive_rebounds_value):
                continue
            offensive_rebounds = float(offensive_rebounds_value)
            if identifier.turnovers_column is None:
                continue
            turnovers_value = row[identifier.turnovers_column]
            if pd.isnull(turnovers_value):
                continue
            turnovers = float(turnovers_value)
            offensive_efficiency_column = DELIMITER.join(
                [identifier.column_prefix, OFFENSIVE_EFFICIENCY_COLUMN]
            )
            row[offensive_efficiency_column] = (field_goals + assists) / (
                field_goals_attempted - offensive_rebounds + assists + turnovers
            )
            if offensive_efficiency_column not in identifier.feature_columns:
                identifier.feature_columns.append(offensive_efficiency_column)
        return row

    return df.progress_apply(record_offensive_efficiency, axis=1)  # type: ignore



"""Tests for the process function."""
import datetime
import os
import tempfile
import unittest

import pandas as pd

from sportsfeatures.process import process
from sportsfeatures.identifier import Identifier
from sportsfeatures.entity_type import EntityType


class TestProcess(unittest.TestCase):

    def test_process(self):
        current_dir = os.getcwd()
        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                team_0_column_prefix = "teams/0"
                team_1_column_prefix = "teams/1"
                dt_column = "dt"
                team_0_id_column = team_0_column_prefix + "/id"
                team_0_kicks = team_0_column_prefix + "/kicks"
                team_0_points_column = team_0_column_prefix + "/points"
                team_0_field_goals_column = team_0_column_prefix + "/field_goals"
                team_0_assists_column = team_0_column_prefix + "/assists"
                team_0_field_goals_attempted_column = team_0_column_prefix + "/field_goals_attempted"
                team_0_offensive_rebounds_column = team_0_column_prefix + "/offensive_rebounds"
                team_0_turnovers_column = team_0_column_prefix + "/turnovers"
                team_1_id_column = team_1_column_prefix + "/id"
                team_1_kicks = team_1_column_prefix + "/kicks"
                team_1_points_column = team_1_column_prefix + "/points"
                team_1_field_goals_column = team_1_column_prefix + "/field_goals"
                team_1_assists_column = team_1_column_prefix + "/assists"
                team_1_field_goals_attempted_column = team_1_column_prefix + "/field_goals_attempted"
                team_1_offensive_rebounds_column = team_1_column_prefix + "/offensive_rebounds"
                team_1_turnovers_column = team_1_column_prefix + "/turnovers"
                df = pd.DataFrame(data={
                    dt_column: [datetime.datetime(2022, 1, 1), datetime.datetime(2022, 1, 2), datetime.datetime(2022, 1, 3)],
                    team_0_id_column: ["0", "1", "0"],
                    team_0_kicks: [10, 20, 30],
                    team_0_points_column: [50, 100, 150],
                    team_0_field_goals_column: [12, 24, 36],
                    team_0_assists_column: [10, 20, 30],
                    team_0_field_goals_attempted_column: [20, 40, 60],
                    team_0_offensive_rebounds_column: [30, 60, 90],
                    team_0_turnovers_column: [10.0, 20.0, 30.0],
                    team_1_id_column: ["1", "0", "1"],
                    team_1_kicks: [20, 40, 60],
                    team_1_points_column: [60, 120, 180],
                    team_1_field_goals_column: [30, 60, 90],
                    team_1_assists_column: [20, 40, 60],
                    team_1_field_goals_attempted_column: [80, 160, 240],
                    team_1_offensive_rebounds_column: [90, 180, 270],
                    team_1_turnovers_column: [10.0, 20.0, 30.0],
                })
                identifiers = [
                    Identifier(
                        EntityType.TEAM,
                        team_0_id_column,
                        [team_0_kicks],
                        team_0_column_prefix,
                        points_column=team_0_points_column,
                        field_goals_column=team_0_field_goals_column,
                        assists_column=team_0_assists_column,
                        field_goals_attempted_column=team_0_field_goals_attempted_column,
                        offensive_rebounds_column=team_0_offensive_rebounds_column,
                        turnovers_column=team_0_turnovers_column,
                    ),
                    Identifier(
                        EntityType.TEAM,
                        team_1_id_column,
                        [team_1_kicks],
                        team_1_column_prefix,
                        points_column=team_1_points_column,
                        field_goals_column=team_1_field_goals_column,
                        assists_column=team_1_assists_column,
                        field_goals_attempted_column=team_1_field_goals_attempted_column,
                        offensive_rebounds_column=team_1_offensive_rebounds_column,
                        turnovers_column=team_1_turnovers_column,
                    ),
                ]
                df = process(df, dt_column, identifiers, [datetime.timedelta(days=365), None])
                print(df)
                print(df.columns.values)
        finally:
            os.chdir(current_dir)



"""Process a dataframe for its datetime information."""

import datetime
import random

import pandas as pd
from feature_engine.datetime import DatetimeFeatures
from pandas.api.types import is_datetime64_any_dtype


# https://gist.github.com/rg3915/db907d7455a4949dbe69
def _gen_datetime(
    min_year=1900, max_year=datetime.datetime.now().year
) -> datetime.datetime:
    """Generate a datetime in format yyyy-mm-dd hh:mm:ss.000000."""
    start = datetime.datetime(min_year, 1, 1, 00, 00, 00)
    years = max_year - min_year + 1
    end = start + datetime.timedelta(days=365 * years)
    return start + (end - start) * random.random()


def datetime_process(df: pd.DataFrame) -> pd.DataFrame:
    """Process datetime features."""
    dtf = DatetimeFeatures(
        features_to_extract="all",
        missing_values="ignore",
        drop_original=False,
        utc=True,
    )
    dt_columns = [x for x in df.columns if is_datetime64_any_dtype(x)]
    replace_dt = None
    while replace_dt is None:
        random_dt = _gen_datetime()
        found = False
        for column in dt_columns:
            unique_dts = df[column].unique()
            if random_dt in unique_dts:
                found = True
                break
        if not found:
            replace_dt = random_dt
    df = df.replace({pd.NaT: replace_dt})
    for dt_column in dt_columns:
        df[dt_column] = pd.to_datetime(dt_column, utc=True, errors="coerce")
    df = dtf.fit_transform(df)

    def fix_dummy_date(row: pd.Series) -> pd.Series:
        for dt_column in dt_columns:
            if row[dt_column] != replace_dt:
                continue
            row[dt_column] = pd.NaT
            for column in df.columns.values:
                if column.startswith(str(dt_column)):
                    row[column] = None
        return row

    return df.apply(fix_dummy_date, axis=1)



"""The main process function."""

import datetime

import pandas as pd

from .datetime_process import datetime_process
from .datetimesub_process import datetimesub_process
from .identifier import Identifier
from .offensive_efficiency_process import offensive_efficiency_process
from .remove_process import remove_process
from .skill_process import skill_process
from .timeseries_process import timeseries_process


def process(
    df: pd.DataFrame,
    dt_column: str,
    identifiers: list[Identifier],
    windows: list[datetime.timedelta | None],
) -> pd.DataFrame:
    """Process the dataframe for sports features."""
    df = skill_process(df, dt_column, identifiers, windows)
    df = offensive_efficiency_process(df, identifiers)
    df = datetimesub_process(df, dt_column, identifiers)
    df = timeseries_process(df, identifiers, windows, dt_column)
    df = datetime_process(df)
    df = remove_process(df, identifiers)
    return df



"""Tests for the timeseries process function."""
import datetime
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from sportsfeatures.timeseries_process import _extract_identifier_timeseries, _process_identifier_ts
from sportsfeatures.identifier import Identifier
from sportsfeatures.entity_type import EntityType


class TestTimeseriesProcess(unittest.TestCase):

    def test_extract_identifier_timeseries(self):
        team_0_column_prefix = "teams/0"
        team_1_column_prefix = "teams/1"
        dt_column = "dt"
        team_0_id_column = team_0_column_prefix + "/id"
        team_0_kicks = team_0_column_prefix + "/kicks"
        team_0_id = "0"
        team_1_id_column = team_1_column_prefix + "/id"
        team_1_kicks = team_1_column_prefix + "/kicks"
        team_1_id = "1"
        df = pd.DataFrame(data={
            dt_column: [datetime.datetime(2022, 1, 1), datetime.datetime(2022, 1, 2), datetime.datetime(2022, 1, 3)],
            team_0_id_column: [team_0_id, team_1_id, team_0_id],
            team_0_kicks: [10.0, 20.0, 30.0],
            team_1_id_column: [team_1_id, team_0_id, team_1_id],
            team_1_kicks: [20.0, 40.0, 60.0],
        })
        identifiers = [
            Identifier(
                EntityType.TEAM,
                team_0_id_column,
                [team_0_kicks],
                team_0_column_prefix,
            ),
            Identifier(
                EntityType.TEAM,
                team_1_id_column,
                [team_1_kicks],
                team_1_column_prefix,
            ),
        ]
        ts_dfs = _extract_identifier_timeseries(df, identifiers, dt_column)
        expected_ts_dfs = {
            "_".join([EntityType.TEAM, team_0_id]): pd.DataFrame(data={
                dt_column: df[dt_column],
                "/kicks": [10.0, 40.0, 30.0],
            }),
            "_".join([EntityType.TEAM, team_1_id]): pd.DataFrame(data={
                dt_column: df[dt_column],
                "/kicks": [20.0, 20.0, 60.0],
            })
        }
        for key, value in expected_ts_dfs.items():
            compare_value = ts_dfs[key]
            assert_frame_equal(value, compare_value)

    def test_process_identifier_ts(self):
        dt_column = "dt"
        identifier_ts = {
            "team_0": pd.DataFrame(data={
                "/kicks": [10.0, 20.0, 30.0],
                dt_column: [datetime.datetime(2022, 1, 1), datetime.datetime(2022, 1, 2), datetime.datetime(2022, 1, 3)],
            })
        }
        identifier_ts = _process_identifier_ts(identifier_ts, [datetime.timedelta(days=20), None], dt_column)
        test_df = pd.DataFrame(data={
            "/kicks_sin": [None, 0.8660254037844387, -0.8660254037844384],
            "/kicks_cos": [None, -0.49999999999999983, -0.5000000000000004],
            "/kicks_count_20days": [None, 1.0, 2.0],
            "/kicks_sum_20days": [None, 10.0, 30.0],
            "/kicks_mean_20days": [None, 10.0, 15.0],
            "/kicks_median_20days": [None, 10.0, 15.0],
            "/kicks_var_20days": [None, None, 50.0],
            "/kicks_std_20days": [None, None, 7.071068],
            "/kicks_min_20days": [None, 10.0, 10.0],
            "/kicks_max_20days": [None, 10.0, 20.0],
            "/kicks_skew_20days": [None, None, None],
            "/kicks_kurt_20days": [None, None, None],
            "/kicks_sem_20days": [None, None, 7.071068],
            "/kicks_rank_20days": [None, 1.0, 2.0],
            "/kicks_count_all": [None, 1.0, 2.0],
            "/kicks_sum_all": [None, 10.0, 30.0],
            "/kicks_mean_all": [None, 10.0, 15.0],
            "/kicks_median_all": [None, 10.0, 15.0],
            "/kicks_var_all": [None, None, 50.0],
            "/kicks_std_all": [None, None, 7.071068],
            "/kicks_min_all": [None, 10.0, 10.0],
            "/kicks_max_all": [None, 10.0, 20.0],
            "/kicks_skew_all": [None, None, None],
            "/kicks_kurt_all": [None, None, None],
            "/kicks_sem_all": [None, None, 7.071068],
            "/kicks_rank_all": [None, 1.0, 2.0],
            "/kicks_lag_1": [None, 10.0, 20.0],
            "/kicks_lag_2": [None, None, 10.0],
            "/kicks_lag_4": [None, None, None],
            "/kicks_lag_8": [None, None, None],
        })
        test_df["/kicks_skew_20days"] = test_df["/kicks_skew_20days"].astype(float)
        test_df["/kicks_kurt_20days"] = test_df["/kicks_kurt_20days"].astype(float)
        test_df["/kicks_skew_all"] = test_df["/kicks_skew_all"].astype(float)
        test_df["/kicks_kurt_all"] = test_df["/kicks_kurt_all"].astype(float)
        test_df["/kicks_lag_4"] = test_df["/kicks_lag_4"].astype(float)
        test_df["/kicks_lag_8"] = test_df["/kicks_lag_8"].astype(float)
        assert_frame_equal(identifier_ts["team_0"], test_df)

    def test_nan_in_timeseries_process(self):
        dt_column = "dt"
        identifier_ts = {
            "team_0": pd.DataFrame(data={
                "/kicks": [10.0, 20.0, 30.0, None, None, 40.0],
                dt_column: [
                    datetime.datetime(2022, 1, 1),
                    datetime.datetime(2022, 1, 2),
                    datetime.datetime(2022, 1, 3),
                    datetime.datetime(2022, 1, 4),
                    datetime.datetime(2022, 1, 5),
                    datetime.datetime(2022, 1, 6),
                ],
            })
        }
        identifier_ts = _process_identifier_ts(identifier_ts, [datetime.timedelta(days=20), None], dt_column)
        print(identifier_ts)

    def test_all_nan_in_timeseries_process(self):
        dt_column = "dt"
        identifier_ts = {
            "team_0": pd.DataFrame(data={
                "/kicks": ["all", "all", "all", "all", "all", "all"],
                dt_column: [
                    datetime.datetime(2022, 1, 1),
                    datetime.datetime(2022, 1, 2),
                    datetime.datetime(2022, 1, 3),
                    datetime.datetime(2022, 1, 4),
                    datetime.datetime(2022, 1, 5),
                    datetime.datetime(2022, 1, 6),
                ],
            })
        }
        identifier_ts = _process_identifier_ts(identifier_ts, [datetime.timedelta(days=20), None], dt_column)
        print(identifier_ts)



"""A class for describing a ratings match."""

# pylint: disable=too-few-public-methods

import datetime

from .team import Team


class Match:
    """Store information about a match."""

    def __init__(self, teams: list[Team], dt: datetime.datetime):
        self.teams = teams
        self.dt = dt



"""Processing for time series features."""

import datetime
import logging

import pandas as pd
from feature_engine.creation import CyclicalFeatures
from tqdm import tqdm

from .columns import DELIMITER
from .identifier import Identifier


def _extract_identifier_timeseries(
    df: pd.DataFrame, identifiers: list[Identifier], dt_column: str
) -> dict[str, pd.DataFrame]:
    tqdm.pandas(desc="Timeseries Progress")
    identifier_ts: dict[str, pd.DataFrame] = {}

    def record_timeseries_features(row: pd.Series) -> pd.Series:
        nonlocal identifier_ts

        for identifier in identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            df = identifier_ts.get(key, pd.DataFrame())
            df.loc[row.name, dt_column] = row[dt_column]  # type: ignore
            for feature_column in identifier.feature_columns:
                if feature_column not in row:
                    continue
                value = row[feature_column]
                if pd.isnull(value):
                    continue
                column = feature_column[len(identifier.column_prefix) :]
                if column not in df:
                    df[column] = None
                df.loc[row.name, column] = value  # type: ignore
            identifier_ts[key] = df.infer_objects()

        return row

    df.progress_apply(record_timeseries_features, axis=1)  # type: ignore
    return identifier_ts


def _process_identifier_ts(
    identifier_ts: dict[str, pd.DataFrame],
    windows: list[datetime.timedelta | None],
    dt_column: str,
) -> dict[str, pd.DataFrame]:
    # pylint: disable=too-many-locals
    for identifier_id in tqdm(identifier_ts):
        identifier_df = identifier_ts[identifier_id]
        original_identifier_df = identifier_df.copy()
        drop_columns = original_identifier_df.columns.values
        cf = CyclicalFeatures()
        try:
            identifier_df = cf.fit_transform(identifier_df.fillna(0.0))
        except TypeError as exc:
            logging.warning(
                "Failed to fit cyclical transforms for identifier %s: %s",
                identifier_id,
                str(exc),
            )
            continue
        for window in windows + [1, 2, 4, 8]:
            if isinstance(window, int):
                lag_df = (
                    original_identifier_df.shift(window - 1)
                    if window != 1
                    else original_identifier_df
                )
                for column in original_identifier_df.columns.values:
                    if column == dt_column:
                        continue
                    feature_column = DELIMITER.join([column, "lag", str(window)])
                    identifier_df[feature_column] = lag_df[column]
            else:
                window_df = (
                    identifier_df.rolling(window, on=dt_column)
                    if window is not None
                    else identifier_df.expanding()
                )
                window_col = str(window.days) + "days" if window is not None else "all"
                for column in original_identifier_df.columns.values:
                    if column == dt_column:
                        continue
                    count_column = DELIMITER.join([column, "count", window_col])  # type: ignore
                    sum_column = DELIMITER.join([column, "sum", window_col])  # type: ignore
                    mean_column = DELIMITER.join([column, "mean", window_col])  # type: ignore
                    median_column = DELIMITER.join([column, "median", window_col])  # type: ignore
                    var_column = DELIMITER.join([column, "var", window_col])  # type: ignore
                    std_column = DELIMITER.join([column, "std", window_col])  # type: ignore
                    min_column = DELIMITER.join([column, "min", window_col])  # type: ignore
                    max_column = DELIMITER.join([column, "max", window_col])  # type: ignore
                    skew_column = DELIMITER.join([column, "skew", window_col])  # type: ignore
                    kurt_column = DELIMITER.join([column, "kurt", window_col])  # type: ignore
                    sem_column = DELIMITER.join([column, "sem", window_col])  # type: ignore
                    rank_column = DELIMITER.join([column, "rank", window_col])  # type: ignore
                    identifier_df[count_column] = window_df[column].count()
                    identifier_df[sum_column] = window_df[column].sum()
                    identifier_df[mean_column] = window_df[column].mean()
                    identifier_df[median_column] = window_df[column].median()
                    identifier_df[var_column] = window_df[column].var()
                    identifier_df[std_column] = window_df[column].std()
                    identifier_df[min_column] = window_df[column].min()
                    identifier_df[max_column] = window_df[column].max()
                    identifier_df[skew_column] = window_df[column].skew()
                    identifier_df[kurt_column] = window_df[column].kurt()
                    identifier_df[sem_column] = window_df[column].sem()
                    identifier_df[rank_column] = window_df[column].rank()
        identifier_ts[identifier_id] = identifier_df.shift(1).drop(columns=drop_columns)

    return identifier_ts


def _write_ts_features(
    df: pd.DataFrame,
    identifier_ts: dict[str, pd.DataFrame],
    identifiers: list[Identifier],
) -> pd.DataFrame:
    def write_timeseries_features(row: pd.Series) -> pd.Series:
        nonlocal identifier_ts
        nonlocal identifiers

        for identifier in identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            identifier_df = identifier_ts[key]
            identifier_row = identifier_df.iloc[0]
            for column in identifier_df.columns.values:
                new_column = identifier.column_prefix + column
                row[new_column] = identifier_row[column]
            identifier_ts[key] = identifier_df.iloc[1:]

        return row

    return df.progress_apply(write_timeseries_features, axis=1)  # type: ignore


def timeseries_process(
    df: pd.DataFrame,
    identifiers: list[Identifier],
    windows: list[datetime.timedelta | None],
    dt_column: str,
) -> pd.DataFrame:
    """Process a dataframe for its timeseries features."""
    # pylint: disable=too-many-locals,consider-using-dict-items,too-many-statements,duplicate-code
    tqdm.pandas(desc="Progress")
    identifier_ts: dict[str, pd.DataFrame] = _extract_identifier_timeseries(
        df, identifiers, dt_column
    )
    identifier_ts = _process_identifier_ts(identifier_ts, windows, dt_column)
    return _write_ts_features(df, identifier_ts, identifiers)



"""A class for handling windowed plackett luce ratings."""

# pylint: disable=too-many-locals,too-many-branches,too-many-statements,too-few-public-methods,too-many-nested-blocks

import datetime

import pandas as pd
from openskill.models import PlackettLuce, PlackettLuceRating

from .identifier import Identifier
from .match import Match
from .team import Team


class WindowedRating:
    """Handles plackett luce ratings based on windows."""

    _team_ratings: dict[str, PlackettLuceRating]
    _player_ratings: dict[str, PlackettLuceRating]
    _matches: list[Match]

    def __init__(self, window: datetime.timedelta | None, dt_column: str):
        self.window = window
        self._dt_column = dt_column
        self._team_model = PlackettLuce()
        self._player_model = PlackettLuce()
        self._team_ratings = {}
        self._player_ratings = {}
        self._matches = []

    def add(
        self,
        row: pd.Series,
        team_identifiers: list[Identifier],
        player_identifiers: list[Identifier],
    ) -> tuple[
        dict[str, tuple[PlackettLuceRating, int, float]],
        dict[str, tuple[PlackettLuceRating, int, float]],
    ]:
        """Add a new row to the windowed rating."""
        teams = []
        for team_identifier in team_identifiers:
            if team_identifier.column not in row:
                continue
            team_id = row[team_identifier.column]
            if pd.isnull(team_id):
                continue
            if team_id not in self._team_ratings:
                self._team_ratings[team_id] = self._team_model.rating(name=team_id)

            team_player_ids = []
            for player_identifier in player_identifiers:
                if player_identifier.team_identifier_column is None:
                    continue
                if player_identifier.team_identifier_column not in row:
                    continue
                player_team_id = row[player_identifier.team_identifier_column]
                if pd.isnull(player_team_id):
                    continue
                if player_team_id != team_id:
                    continue
                if player_identifier.column not in row:
                    continue
                player_id = row[player_identifier.column]
                if pd.isnull(player_id):
                    continue
                team_player_ids.append(player_id)
                if player_id not in self._player_ratings:
                    self._player_ratings[player_id] = self._player_model.rating(
                        name=player_id
                    )

            points = None
            if (
                team_identifier.points_column is not None
                and team_identifier.points_column in row
            ):
                points = row[team_identifier.points_column]
            teams.append(Team(team_player_ids, points, team_id))

        match = Match(teams, row[self._dt_column].to_pydatetime())

        if self.window is not None:
            # Remove older matches and reverse their results.
            oldest_dt = match.dt - self.window
            remove_matches = [x for x in self._matches if x.dt < oldest_dt]
            for remove_match in remove_matches:
                scores: list[float] | None = [
                    x.points
                    for x in reversed(remove_match.teams)
                    if x.points is not None
                ]
                if not scores:
                    scores = None
                if len(remove_match.teams) >= 2:
                    team_ratings: list[list[PlackettLuceRating]] = [
                        [self._team_ratings[x.identifier]] for x in remove_match.teams
                    ]
                    team_ratings = self._team_model.rate(
                        team_ratings,
                        scores=scores,
                    )
                    for team_rating in team_ratings:
                        self._team_ratings[str(team_rating[0].name)] = team_rating[0]
                    if all(x.players for x in remove_match.teams):
                        players = self._player_model.rate(
                            [
                                [self._player_ratings[y] for y in x.players]
                                for x in remove_match.teams
                            ],
                            scores=scores,
                        )
                        for player_rating in players:
                            for player_subrating in player_rating:
                                self._player_ratings[str(player_subrating.name)] = (
                                    player_subrating
                                )
            self._matches = self._matches[len(remove_matches) :]

        # Find the results
        team_result = {}
        player_result = {}
        if len(match.teams) >= 2:
            team_rank = self._team_model.predict_rank(
                [[self._team_ratings[x.identifier]] for x in match.teams]
            )
            team_result = {
                x.identifier: (
                    self._team_ratings[x.identifier],
                    team_rank[count][0],
                    team_rank[count][1],
                )
                for count, x in enumerate(match.teams)
            }
            if all(x.players for x in match.teams):
                player_rank = self._player_model.predict_rank(
                    [[self._player_ratings[y] for y in x.players] for x in match.teams]
                )
                for count, team in enumerate(match.teams):
                    for player in team.players:
                        player_result[player] = (
                            self._player_ratings[player],
                            player_rank[count][0],
                            player_rank[count][1],
                        )

        # Record the new match results
        scores = [x.points for x in match.teams if x.points is not None]
        if not scores:
            scores = None
        if len(match.teams) >= 2:
            team_ratings = self._team_model.rate(
                [[self._team_ratings[x.identifier]] for x in match.teams], scores=scores
            )
            for team_rating in team_ratings:
                self._team_ratings[str(team_rating[0])] = team_rating[0]
            if all(x.players for x in match.teams):
                player_ratings: list[list[PlackettLuceRating]] = (
                    self._player_model.rate(
                        [
                            [self._player_ratings[y] for y in x.players]
                            for x in match.teams
                        ],
                        scores=scores,
                    )
                )
                for player_rating in player_ratings:
                    for player_subrating in player_rating:
                        self._player_ratings[str(player_subrating.name)] = (
                            player_subrating
                        )
        self._matches.append(match)

        return team_result, player_result



"""The process for removing lookahead features."""

import pandas as pd

from .identifier import Identifier


def remove_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Remove the features from the dataframe."""
    drop_columns: set[str] = set()
    for identifier in identifiers:
        for feature_col in identifier.feature_columns:
            drop_columns.add(feature_col)
    return df.drop(columns=list(drop_columns))



"""Caching functionality for sportsfeatures."""

import os

from . import __VERSION__

_SPORTS_FEATURES_CACHE_FOLDER = ".sportsfeatures_" + __VERSION__


def sportsfeatures_cache_folder() -> str:
    """Return a valid cache folder."""
    if not os.path.exists(_SPORTS_FEATURES_CACHE_FOLDER):
        try:
            os.mkdir(_SPORTS_FEATURES_CACHE_FOLDER)
        except FileExistsError:
            pass
    return _SPORTS_FEATURES_CACHE_FOLDER



"""Process the current dataframe by adding skill features."""

# pylint: disable=duplicate-code

import datetime

import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .entity_type import EntityType
from .identifier import Identifier
from .windowed_rating import WindowedRating

SKILL_COLUMN_PREFIX = "skill"
SKILL_MU_COLUMN = "mu"
SKILL_SIGMA_COLUMN = "sigma"
SKILL_RANKING_COLUMN = "ranking"
SKILL_PROBABILITY_COLUMN = "probability"
TIME_SLICE_ALL = "all"


def skill_process(
    df: pd.DataFrame,
    dt_column: str,
    identifiers: list[Identifier],
    windows: list[datetime.timedelta | None],
) -> pd.DataFrame:
    """Add skill features to the dataframe."""
    tqdm.pandas(desc="Skill Features")

    team_identifiers = [x for x in identifiers if x.entity_type == EntityType.TEAM]
    player_identifiers = [x for x in identifiers if x.entity_type == EntityType.PLAYER]
    rating_windows = [WindowedRating(x, dt_column) for x in windows]

    def calculate_skills(row: pd.Series) -> pd.Series:
        nonlocal rating_windows
        nonlocal team_identifiers
        nonlocal player_identifiers

        for rating_window in rating_windows:
            window_id = (
                TIME_SLICE_ALL
                if rating_window.window is None
                else f"window{rating_window.window.days}"
            )
            team_result, player_result = rating_window.add(
                row, team_identifiers, player_identifiers
            )
            for team_identifier in team_identifiers:
                if team_identifier.column not in row:
                    continue
                team_id = row[team_identifier.column]
                if pd.isnull(team_id):
                    continue
                if team_id in team_result:
                    rating, ranking, prob = team_result[team_id]
                    window_prefix = DELIMITER.join(
                        [team_identifier.column_prefix, SKILL_COLUMN_PREFIX, window_id]
                    )
                    row[DELIMITER.join([window_prefix, SKILL_MU_COLUMN])] = rating.mu
                    row[DELIMITER.join([window_prefix, SKILL_SIGMA_COLUMN])] = (
                        rating.sigma
                    )
                    row[DELIMITER.join([window_prefix, SKILL_RANKING_COLUMN])] = ranking
                    row[DELIMITER.join([window_prefix, SKILL_PROBABILITY_COLUMN])] = (
                        prob
                    )
                for player_identifier in player_identifiers:
                    if player_identifier.column not in row:
                        continue
                    player_id = row[player_identifier.column]
                    if pd.isnull(player_id):
                        continue
                    if player_id in player_result:
                        rating, ranking, prob = team_result[team_id]
                        window_prefix = DELIMITER.join(
                            [
                                player_identifier.column_prefix,
                                SKILL_COLUMN_PREFIX,
                                window_id,
                            ]
                        )
                        row[DELIMITER.join([window_prefix, SKILL_MU_COLUMN])] = (
                            rating.mu
                        )
                        row[DELIMITER.join([window_prefix, SKILL_SIGMA_COLUMN])] = (
                            rating.sigma
                        )
                        row[DELIMITER.join([window_prefix, SKILL_RANKING_COLUMN])] = (
                            ranking
                        )
                        row[
                            DELIMITER.join([window_prefix, SKILL_PROBABILITY_COLUMN])
                        ] = prob

        return row

    return df.progress_apply(calculate_skills, axis=1)  # type: ignore
